from typing import Union
from urllib import parse

import nonebot
from aiohttp.client_exceptions import ContentTypeError

from .mwapi import Mwapi
from .exception import MediaWikiException, MediaWikiGeoCoordError, HTTPTimeoutError, PageError


class Wiki:
    def __init__(self, api_url: str, fallback_url: str):
        self.__url = fallback_url
        self.__api_url = api_url

    async def get_from_api(self, title: str, is_template: bool) -> tuple[bool, Union[str, list]]:
        """第一个值表示是否为特殊类型（目前是列表选择），第二个是查询结果（正常情况是字符串，特殊情况是列表）"""
        if is_template:
            title = "Template:" + title
        if self.__api_url == '':
            title, anchor = Mwapi.handle_anchor_pre_process(title)
            special, url = await self.url_parse(title)
            url = Mwapi.handle_anchor_post_process(url, anchor)
            result = f"标题：{title}\n链接：{url}"
            return False, result

        try:
            mediawiki = Mwapi(url=self.__url, api_url=self.__api_url)
            result_dict = await mediawiki.get_page_info(title)
        except HTTPTimeoutError:
            exception = "连接超时"
        except (MediaWikiException, MediaWikiGeoCoordError, ContentTypeError) as e:  # ContentTypeError：非json内容
            exception = "Api调用出错"
            nonebot.logger.warning(e)
        except PageError:
            exception = "未找到页面"
        if "exception" in locals():
            special, result = await self.url_parse(title)
            result_dict = {
                "exception": True,
                "title": title,
                "url": result,
                "notes": exception,
            }

        if result_dict["exception"]:
            result = f"错误：{result_dict['notes']}\n" \
                     f"由条目名直接生成的链接：\n" \
                     f"标题：{result_dict['title']}\n链接：{result_dict['url']}"
            return False, result
        if result_dict["interwiki"]:
            result = f"跨wiki链接：{result_dict['title']}\n" \
                     f"链接：{result_dict['url']}"
            return False, result
        if result_dict["missing"] or result_dict["disambiguation"]:
            result = result_dict["notes"]
            result.append((result_dict["missing"], result_dict["title"]))  # 最后一个值保存是缺失还是重定向，以及目标页面标题
            return True, result
        if result_dict["redirected"]:
            result = f"重定向：{result_dict['from_title']} → {result_dict['title']}\n" \
                     f"链接：{result_dict['url']}"
            return False, result

        result = f"标题：{result_dict['title']}\n链接：{result_dict['url']}"
        return False, result

    async def url_parse(self, title: str) -> tuple:
        result = f"{self.__url}/{parse.quote(title)}"
        return False, result



import re
from typing import Optional

from nonebot import Type
from nonebot.adapters.onebot.v11 import Bot, utils, GroupMessageEvent, Message
from nonebot.internal.matcher import Matcher

from .config import Config
from .data_source import Wiki
from .exception import NoDefaultPrefixException, NoSuchPrefixException

__all__ = ['wiki_process', 'wiki_parse']

# 匹配模板
ARTICLE = r"\[\[(.*?)\]\]"
TEMPLATE = r"\{\{(.*?)\}\}"

# 标题列表
titles = []


async def wiki_process(bot: Bot, event: GroupMessageEvent, wiki: Type[Matcher], is_template: bool = False):
    global titles
    message = str(event.message).strip()
    is_raw = False  # 回复0的话这里应该变成True,否则会掉进循环（
    if message.isdigit():
        if 0 <= int(message) < len(titles) - 1:
            event.message = Message(f"[[{titles[-1]}:{titles[int(message)]}]]")
            is_template = False  # 否则会导致下面发生误判，导致无法匹配
            is_raw = bool(1 - bool(int(message)))
        else:
            return
    try:
        special, result = await wiki_parse(ARTICLE if not is_template else TEMPLATE, is_template, is_raw, bot, event)
    except TypeError:
        return  # 这里应当是bot返回选择列表后，用户没有做出有效选择，无害
    if special:
        titles = result[:-1]
        titles.insert(0, result[-1][1])
        titles.append(result[-1][2])
        title_list = '\n'.join([f'{i + 1}.{result[i]}' for i in range(len(result) - 1)])  # 最后一个元素是特殊标记
        msg = f"{f'页面“{result[-1][1]}”不存在，下面是推荐的结果' if result[-1][0] else f'页面“{result[-1][1]}”是消歧义页面'}，" \
              f"请回复数字来选择你想要查询的条目，或者回复0来根据原标题直接生成链接：\n" \
              f"{title_list}"
        await wiki.reject(msg)
    else:
        await bot.send(event, result)
        titles = []


async def wiki_parse(pattern: str, is_template: bool, is_raw: bool, bot: Bot, event: GroupMessageEvent) \
        -> Optional[tuple]:
    msg = str(event.message).strip()
    msg = utils.unescape(msg)  # 将消息处理为正常格式，以防搜索出错
    temp_config: Config = Config(event.group_id)
    tmp_titles = re.findall(pattern, msg)
    for title in tmp_titles:
        title = str(title)
        prefix = re.match(r'\w+:|\w+：', title)
        if not prefix:
            prefix = ''
        else:
            prefix = prefix.group(0).lower().rstrip(":：")  # 删掉右侧冒号以进行匹配
            if prefix in temp_config.prefixes:
                title = re.sub(f"{prefix}:|{prefix}：", '', title, count=1, flags=re.I)  # 去除标题左侧的前缀
            else:
                prefix = ''  # 如果不在前缀列表里，视为名字空间标识，回落到默认前缀

        try:
            if title is None or title.strip() == "":
                continue
            wiki_api = temp_config.get_from_prefix(prefix)[0]
            wiki_url = temp_config.get_from_prefix(prefix)[1]

            wiki_object = Wiki(wiki_api, wiki_url)
            if not is_raw:
                special, result = await wiki_object.get_from_api(title, is_template)
            else:
                special, url = await wiki_object.url_parse(title)
                result = f"标题：{title}\n链接：{url}"

            if special:
                result[-1] = list(result[-1])
                result[-1].append(prefix)  # 补充前缀，防止一会查的时候回落到默认wiki

        except NoDefaultPrefixException:
            special = False
            result = "没有找到默认前缀，请群管或bot管理员先设置默认前缀"
        except NoSuchPrefixException:
            special = False
            result = "指定的默认前缀对应的wiki不存在，请管理员检查设置"

        return special, result



import sys

'''
代码来自 pymediawiki 库（以MIT许可证开源），并根据bot的实际需要做了一些修改
该库的Github地址：https://github.com/barrust/mediawiki
许可证：https://github.com/barrust/mediawiki/blob/master/LICENSE
'''


def str_or_unicode(text):
    """ handle python 3 unicode """
    encoding = sys.stdout.encoding
    return text.encode(encoding).decode(encoding)



import re
from typing import Type

from nonebot import on_command, get_driver
from nonebot.adapters.onebot.v11 import GroupMessageEvent, Bot, MessageEvent
from nonebot.adapters.onebot.v11.permission import GROUP_OWNER, GROUP_ADMIN
from nonebot.permission import SUPERUSER
from nonebot.typing import T_State
from nonebot.internal.matcher import Matcher

from .config import Config
from .mwapi import Mwapi

QUIT_LIST = ["取消", "quit", "退出"]
BotConfig = get_driver().config

add_wiki = on_command("wiki.add", permission=SUPERUSER | GROUP_ADMIN | GROUP_OWNER)


@add_wiki.handle()
async def _add_wiki(bot: Bot, event: MessageEvent, state: T_State):
    msg = str(event.message).strip().removeprefix("wiki.add")
    state["global"] = msg == ".global"

    if state["global"] and str(event.user_id) not in BotConfig.superusers:
        await add_wiki.finish()


@add_wiki.got("prefix", "请回复前缀，或回复“取消”以退出")
async def _add_wiki_prefix(bot: Bot, event: MessageEvent, state: T_State):
    prefix = str(state["prefix"]).strip().lower()
    reserved = ["talk", "user", "user talk", "project", "project talk", "file", "file talk", "mediawiki",
                "mediawiki talk", "template", "template talk", "help", "help talk", "category", "category talk",
                "special", "media", "t", "u"]

    if prefix in QUIT_LIST:
        await add_wiki.finish("OK")
    elif prefix in reserved:
        await add_wiki.reject("前缀位于保留名字空间！请重新输入！")
    elif re.findall(r'\W', prefix):
        await add_wiki.reject("前缀含有非法字符，请重新输入！")
    else:
        state['prefix'] = prefix


@add_wiki.got("api_url", "请输入wiki的api地址，回复“0”跳过")
async def _add_wiki_api_url(bot: Bot, event: MessageEvent, state: T_State):
    api_url = str(state["api_url"]).strip()

    if api_url == "0":
        state["api_url"] = ''
    elif api_url in QUIT_LIST:
        await add_wiki.finish("OK")
    elif not re.match(r'^https?:/{2}\w.+$', api_url):
        await add_wiki.reject("非法url!请重新输入！")
    else:
        api = Mwapi(url='', api_url=api_url)
        if not await api.test_api():
            await add_wiki.reject("无法连接到api，请重新输入！如果确认无误的话，可能是被防火墙拦截，可以回复“0”跳过，或者“取消”来退出")
        state['api_url'] = api_url.strip().rstrip("/")  # 之所以去掉右边的斜杠，是为了防止之后需要进行拼接等操作时因多出来斜杠而出错


@add_wiki.got("url", "请输入wiki的通用url（必填）")
async def _add_wiki_url(bot: Bot, event: MessageEvent, state: T_State):
    url = str(state["url"]).strip()

    if url in QUIT_LIST:
        await add_wiki.finish("OK")
    elif not re.match(r'^https?:/{2}\w.+$', url):
        await add_wiki.reject("非法url！请重新输入！")
    else:
        state['url'] = url.strip().rstrip("/")

    # 进行插入操作
    group_id = event.group_id if isinstance(event, GroupMessageEvent) else 0
    config = Config(group_id=group_id)
    if (state["global"] and config.add_wiki_global(state["prefix"], state["api_url"], state["url"])) \
            or (not state["global"] and config.add_wiki(state["prefix"], state["api_url"], state["url"])):
        await add_wiki.finish(f"添加/编辑Wiki：{state['prefix']}成功！")
    else:
        await add_wiki.finish("呜……出错了……请联系bot管理员进行处理……")


list_wiki = on_command("wiki.list")


@list_wiki.handle()
async def _list_wiki(bot: Bot, event: MessageEvent, state: T_State):
    msg = str(event.message).strip().removeprefix("wiki.list")
    is_global = msg == ".global"

    if is_global:
        config = Config(group_id=0)
        await list_wiki.finish(config.list_data[1])
    elif isinstance(event, GroupMessageEvent):
        config = Config(group_id=event.group_id)
        await list_wiki.finish(config.list_data[0])


del_wiki = on_command("wiki.delete", permission=SUPERUSER | GROUP_ADMIN | GROUP_OWNER)


@del_wiki.handle()
async def _del_wiki(bot: Bot, event: MessageEvent, state: T_State):
    msg = str(event.message).strip().removeprefix("wiki.delete")

    await __check_params(msg=msg, event=event, state=state, matcher=del_wiki)


@del_wiki.got("prefix", "请回复前缀，或回复“取消”以退出")
async def _del_wiki_prefix(bot: Bot, event: MessageEvent, state: T_State):
    prefix, config = await __check_prefix(event=event, state=state, matcher=del_wiki)

    if (state["global"] and config.del_wiki_global(prefix)) or (not state["global"] and config.del_wiki(prefix)):
        await del_wiki.finish("删除成功")
    else:
        await del_wiki.finish("呜……删除失败了……请检查前缀是否有误")


set_default = on_command("wiki.default", permission=SUPERUSER | GROUP_ADMIN | GROUP_OWNER)


@set_default.handle()
async def _set_default(bot: Bot, event: MessageEvent, state: T_State):
    msg = str(event.message).strip().removeprefix("wiki.default")

    await __check_params(msg=msg, event=event, state=state, matcher=set_default)


@set_default.got("prefix", "请输入要设置的前缀，或者回复“取消”退出")
async def set_default_prefix(bot: Bot, event: MessageEvent, state: T_State):
    prefix, config = await __check_prefix(event=event, state=state, matcher=set_default)

    if (state["global"] and config.set_default_global(prefix)) or (not state["global"] and config.set_default(prefix)):
        await del_wiki.finish("设置成功")
    else:
        await del_wiki.finish("呜……设置失败了……请检查前缀是否有误")


# 一些公用的部分
async def __check_params(msg: str, event: MessageEvent, state: T_State, matcher: Type[Matcher]):
    params_list = msg.split(maxsplit=1)

    state["global"] = params_list[0] == ".global" if params_list else False
    if len(params_list) == 2 and state["global"]:
        state["prefix"] = params_list[1]
    elif params_list:
        state["prefix"] = params_list[0]

    if state["global"] and str(event.user_id) not in BotConfig.superusers:
        await matcher.finish()


async def __check_prefix(event: MessageEvent, state: T_State, matcher: Type[Matcher]) -> tuple[str, Config]:
    prefix = str(state['prefix']).strip()
    state['prefix'] = prefix

    if prefix in QUIT_LIST:
        await matcher.finish("OK")
    elif re.findall(r'\W', prefix):
        await matcher.reject("前缀含有非法字符，请重新输入！")

    group_id = event.group_id if isinstance(event, GroupMessageEvent) else 0
    config = Config(group_id=group_id)

    return prefix, config



from nonebot import on_regex, on_command
from nonebot.adapters.onebot.v11 import Bot, GroupMessageEvent, Message
from nonebot.adapters.onebot.v11.permission import GROUP

from . import config_manager
from .config import Config
from .data_source import Wiki
from .exception import NoDefaultPrefixException, NoSuchPrefixException
from .worker import *

# 接入帮助系统
__usage__ = '使用：\n' \
            '快速使用：wiki 前缀:条目名\n' \
            '[[前缀:条目名]] {{前缀:模板名}} ((前缀:条目名))\n' \
            '其中中括号、大括号匹配后会调用api搜索条目/模板名，如果有误，可以使用小括号方式绕过api直接生成链接\n' \
            '前缀由群管和bot超管配置，没有指定前缀或前缀无效时，会回落到默认前缀\n' \
            '配置（带global的是全局命令，仅超管可以使用）：\n' \
            '添加：wiki.add，wiki.add.global\n' \
            '删除：wiki.delete，wiki.delete.global\n' \
            '列表：wiki.list，wiki.list.global\n' \
            '设置默认：wiki.default，wiki.default.global\n' \
            '按提示提供相应参数即可\n' \
            '注意：私聊状态下该插件仅会响应超管的命令，且仅能管理全局wiki\n' \
            '完整文档请前往 https://github.com/KoishiMoe/nonebot-plugin-mediawiki 查看'

__help_version__ = '0.1.2'

__help_plugin_name__ = 'Wiki推送'

# 用于正则匹配的模板字符串
ARTICLE_RAW = r"&#91;&#91;(.*?)&#93;&#93;"  # adapter出于安全原因会把中括号转义，此处用于让事件响应器能正确响应事件
ARTICLE = r"\[\[(.*?)\]\]"
TEMPLATE = r"\{\{(.*?)\}\}"
RAW = r"\(\((.*?)\)\)"

# 响应器
wiki = on_regex(ARTICLE_RAW, permission=GROUP)
wiki_template = on_regex(TEMPLATE, permission=GROUP)
wiki_raw = on_regex(RAW, permission=GROUP)
wiki_quick = on_command("wiki", permission=GROUP, priority=2)  # 设个更低的优先级，省着和设置管理器抢（虽然似乎不设也不会抢2333


@wiki.handle()
async def _wiki(bot: Bot, event: GroupMessageEvent):
    await wiki_process(bot, event, wiki, is_template=False)


@wiki_template.handle()
async def _wiki_template(bot: Bot, event: GroupMessageEvent):
    await wiki_process(bot, event, wiki_template, is_template=True)


@wiki_raw.handle()
async def _wiki_raw(bot: Bot, event: GroupMessageEvent):
    special, result = await wiki_parse(RAW, False, True, bot, event)
    await bot.send(event, result)


@wiki_quick.handle()
async def _wiki_quick(bot: Bot, event: GroupMessageEvent):
    if str(event.message).strip().startswith("wiki"):  # 防止选择式的消息及无关消息被误加括号
        message = str(event.message).strip().removeprefix("wiki").strip()
        event2 = event.copy()  # 浅拷贝一下，省着篡改消息后把其他插件弄懵了～
        event2.message = Message("[[" + message + "]]")
        await wiki_process(bot, event2, wiki_quick, is_template=False)
    else:
        await wiki_process(bot, event, wiki_quick, is_template=False)



from .utilities import str_or_unicode


class MediaWikiBaseException(Exception):
    """ Base MediaWikiException

        Args:
            message: The message of the exception """

    def __init__(self, message):
        self._message = message
        super(MediaWikiBaseException, self).__init__(self.message)

    def __unicode__(self):
        return self.message

    def __str__(self):
        return str_or_unicode(self.__unicode__())

    @property
    def message(self):
        """ str: The MediaWiki exception message """
        return self._message


class MediaWikiException(MediaWikiBaseException):
    """ MediaWiki Exception Class

        Args:
            error (str): The error message that the MediaWiki site returned """

    def __init__(self, error):
        self._error = error
        msg = 'An unknown error occurred: "{0}". Please report it on GitHub!'.format(
            self.error
        )
        super(MediaWikiException, self).__init__(msg)

    @property
    def error(self):
        """ str: The error message that the MediaWiki site returned """
        return self._error


class HTTPTimeoutError(MediaWikiBaseException):
    """ Exception raised when a request to the Mediawiki site times out.

        Args:
            query (str): The query that timed out"""

    def __init__(self, query):
        self._query = query
        msg = (
            'Searching for "{0}" resulted in a timeout. '
            "Try again in a few seconds, and ensure you have rate limiting "
            "set to True."
        ).format(self.query)
        super(HTTPTimeoutError, self).__init__(msg)

    @property
    def query(self):
        """ str: The query that timed out """
        return self._query


class MediaWikiGeoCoordError(MediaWikiBaseException):
    """ Exceptions to handle GeoData exceptions

        Args:
            error (str): Error message from the MediaWiki site related to \
                         GeoCoordinates """

    def __init__(self, error):
        self._error = error
        msg = (
            "GeoData search resulted in the following error: {0}"
            " - Please use valid coordinates or a proper page title."
        ).format(self.error)
        super(MediaWikiGeoCoordError, self).__init__(msg)

    @property
    def error(self):
        """ str: The error that was thrown when pulling GeoCoordinates """
        return self._error


class PageError(MediaWikiBaseException):
    """ Exception raised when no MediaWiki page matched a query

        Args:
            title (str): Title of the page
            pageid (int): MediaWiki page id of the page"""

    def __init__(self, title=None, pageid=None):
        if title:
            self._title = title
            msg = '"{0}" does not match any pages. Try another query!'.format(
                self._title
            )
        elif pageid:
            self._pageid = pageid
            msg = 'Page id "{0}" does not match any pages. Try another id!'.format(
                self._pageid
            )
        else:
            self._title = ""
            msg = '"{0}" does not match any pages. Try another query!'.format(
                self._title
            )
        super(PageError, self).__init__(msg)


class NoDefaultPrefixException(MediaWikiBaseException):
    def __init__(self, group: int = None):
        self._group = group
        msg = f"群{group}没有配置默认的Wiki前缀"
        super(NoDefaultPrefixException, self).__init__(msg)

    @property
    def group(self):
        return self._group


class NoSuchPrefixException(MediaWikiBaseException):
    def __init__(self, group: int = None, prefix: str = None):
        self._group = group
        self._prefix = prefix
        msg = f"群{group}的wiki列表以及全局列表中均不存在前缀{prefix}"
        super(NoSuchPrefixException, self).__init__(msg)

    @property
    def group(self):
        return self._group

    @property
    def prefix(self):
        return self._prefix
